#define OPENSSL_THREAD_DEFINES
#include <openssl/opensslconf.h>
#include <openssl/crypto.h>
#include "mod_auth_mutual.h"

#if defined(OPENSSL_THREADS) && APR_HAS_THREADS

static apr_thread_mutex_t **crypto_locks;
static int num_locks;

static unsigned long thread_id_callback(void)
{
  return (unsigned long)apr_os_thread_current();
}

static void locking_callback(int mode, int n, const char *file, int line)
{
  if (n < num_locks) {
    if (mode & CRYPTO_LOCK) {
      apr_thread_mutex_lock(crypto_locks[n]);
    } else {
      apr_thread_mutex_unlock(crypto_locks[n]);
    }
  } else {
    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, NULL, "mod_auth_mutual: libcrypto lock failed: %d", n);
  }
}

static apr_status_t cleanup_crypto_thread(void *d)
{
  ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, NULL, "mod_auth_mutual: cleanup_crypto_thread");
  CRYPTO_set_locking_callback(NULL);
  CRYPTO_set_id_callback(NULL);
  return APR_SUCCESS;
}

void auth_mutual_init_crypto_thread(apr_pool_t *p, server_rec *s)
{
  int i;

  if (CRYPTO_get_locking_callback() != NULL) {
    /* We share libcrypto with other module?
     * Probably mod_ssl?
     */
    ap_log_error(APLOG_MARK, APLOG_INFO, 0, s,
                 "Mutual: detect libcrypto.so is shared with other modules.");
    return;
  }
  num_locks = CRYPTO_num_locks();
  crypto_locks = apr_pcalloc(p, sizeof(*crypto_locks) * num_locks);
  for (i = 0; i < num_locks; i++) {
    apr_thread_mutex_create(&crypto_locks[i], APR_THREAD_MUTEX_NESTED, p);
  }
  CRYPTO_set_id_callback(thread_id_callback);
  CRYPTO_set_locking_callback(locking_callback);

  apr_pool_cleanup_register(p, NULL, cleanup_crypto_thread, apr_pool_cleanup_null);
}

#else

void auth_mutual_init_crypto_thread(apr_pool_t *p, server_rec *s)
{
}

#endif
