#include "mod_auth_mutual.h"

/*
 * Validate and update nonce table atomically.
 *
 * See pseudo code in 'mod_atuh_mutual.h'.
 */
int auth_mutual_validate_and_update_nonce(auth_mutual_nc_t nc, auth_mutual_nc_rec *rec)
{
  int valid;
  auth_mutual_nc_rec old_rec, new_rec;

  do {
    old_rec.t32 = apr_atomic_read32(&rec->t32);
    
    valid = 0;
    new_rec = old_rec;
    if (new_rec.t16.nc_min <= nc && nc <= AUTH_MUTUAL_NC_MAX) {
      int pos = nc - new_rec.t16.nc_min;

      if (pos >= AUTH_MUTUAL_NC_WND) {
        int shift = pos - AUTH_MUTUAL_NC_WND + 1;
        
        valid = 1;
        if (shift >= 16) {
          new_rec.t16.used_bitmap = 0;
        } else {
          new_rec.t16.used_bitmap >>= shift;
        }
        new_rec.t16.nc_min += shift;
        pos = AUTH_MUTUAL_NC_WND - 1;
      } else {
        if (~new_rec.t16.used_bitmap & (1 << pos))
          valid = 1;
      }
      new_rec.t16.used_bitmap |= 1 << pos;
    }
    if (!valid) return valid;
  } while (apr_atomic_cas32(&rec->t32, new_rec.t32, old_rec.t32) != old_rec.t32);
  return valid;
}

void auth_mutual_init_nonce(auth_mutual_nc_rec *rec)
{
  rec->t16.nc_min = 0;
  rec->t16.used_bitmap = 0;
}

void auth_mutual_invalidate_nonce(auth_mutual_nc_rec *rec)
{
  apr_atomic_set32(&rec->t32, (apr_uint32_t)-1);
}
